#include <stdio.h>
#include <gdk-pixbuf/gdk-pixbuf.h>

int count = 0;
typedef struct
{
    int x, y, width, height;
}AreaUpdatedParams;
void callback_area_updated(GdkPixbufLoader* loader, int x, int y, int width, int height,
    AreaUpdatedParams *data)
{
    count++;
    data->x = x;
    data->y = y;
    data->width = width;
    data->height = height;
}

gboolean loader_write_from_channel(GdkPixbufLoader* loader, GIOChannel *channel, gsize count_bytes)
{
    guchar* buffer = g_malloc(count_bytes);
    gsize bytes_read;

    if(g_io_channel_read_chars(channel, (gchar*)buffer, count_bytes, &bytes_read, NULL) != 
        G_IO_STATUS_NORMAL)
    {
        printf("Cannot read bytes from channel.\n");
        g_free(buffer);
        return FALSE;    
    }
    
    if(!gdk_pixbuf_loader_write(loader, buffer, count_bytes, NULL))
    {
        printf("gdk_pixbuf_loader_write() returned FALSE.\n");
        g_free(buffer);
        return FALSE;
    }
    g_free(buffer);
}

/*
 * Create loader of type 'loader_type', and open file 'filename'.
 * Write  first 'initial_read' bytes from the file to the loader.
 *
 * Then store result of gdk_pixbuf_loader_get_pixbuf(loader) to file 'filename_for_initial_pixbuf'.
 * 
 * Then connect callback function for signal "area-updated" and write next 'test_read' bytes
 * from file to loader.
 * Print count of emitting of the "area-update" signal and parameters of this signal.
 * 
 * Then store result of gdk_pixbuf_loader_get_pixbuf(loader) to file 'filename_for_updated_pixbuf'.
 */
 
int test_signal(const gchar* filename, const gchar* loader_type,
                gsize initial_read, gsize test_read,
                const gchar* filename_for_initial_pixbuf, const gchar* filename_for_updated_pixbuf)
{
    GError* error = NULL;
    
    GIOChannel *channel = g_io_channel_new_file(filename, "r", &error);
    if(channel == NULL)
    {
        printf("Cannot open file '%s' for read image: %s.\n", filename, error->message);
        return 1;
    }
    g_io_channel_set_encoding(channel, NULL, NULL);
    
    GdkPixbufLoader* loader = gdk_pixbuf_loader_new_with_type(loader_type, NULL);
    //the first reading
    loader_write_from_channel(loader, channel, initial_read);
            
    GdkPixbuf* pixbuf = gdk_pixbuf_loader_get_pixbuf(loader);
    
    if(pixbuf == NULL)
    {
        printf("Loader hasn't created pixbuf yet.\n");
        gdk_pixbuf_loader_close(loader, NULL);
        g_object_unref(loader);
        g_io_channel_unref(channel);
        return 1;
    }
    
    gdk_pixbuf_save(pixbuf, filename_for_initial_pixbuf, "bmp", NULL, NULL);
    //connect callback
    AreaUpdatedParams data;
    g_signal_connect(loader, "area-updated", (GCallback)callback_area_updated, &data);
    count = 0;
    //continue read
    loader_write_from_channel(loader, channel, test_read);

    g_io_channel_unref(channel);

    pixbuf = gdk_pixbuf_loader_get_pixbuf(loader);

    if(pixbuf == NULL)
    {
        printf("Loader failed to load pixbuf.\n");
        gdk_pixbuf_loader_close(loader, NULL);
        g_object_unref(loader);

        return 1;
    }
    gdk_pixbuf_save(pixbuf, filename_for_updated_pixbuf, "bmp", NULL, NULL);

    gdk_pixbuf_loader_close(loader, NULL);
    g_object_unref(loader);
    
    if(count == 0)
    {
        printf("\"area-updated\" signal wasn't emitted.\n");
        return 1;
    }
    if(count > 1)
    {
        printf("\"area-updated\" signal was emitted more than once.\n");
        return 1;
    }
    printf("\"area-updated\" signal was emitted once\n");
    printf("with parameters: 'x' = %d, 'y' = %d, 'width' = %d, 'height' = %d.\n",
        data.x, data.y, data.width, data.height);
    printf("Initial pixbuf and pixbuf after update was saved in files '%s' and '%s'.\n",
        filename_for_initial_pixbuf, filename_for_updated_pixbuf);
        
    return 0;
}

guint32 get_pixel(const char* filename, int x, int y)
{
    guchar* colors;
    guint32 pixel;
    GError* error = NULL;    
    GdkPixbuf* pixbuf = gdk_pixbuf_new_from_file(filename, &error);
    if(pixbuf == NULL)
    {
        printf("Cannot open image '%s': %s.\n", filename, error->message);
        return 0;
    }
    
    colors = ((guchar*)gdk_pixbuf_get_pixels(pixbuf)
                + gdk_pixbuf_get_n_channels(pixbuf) * x
                + gdk_pixbuf_get_rowstride(pixbuf) * y);
    
    pixel = (colors[0] << 24) | (colors[1] << 16) | (colors[2] << 8);
    if(gdk_pixbuf_get_n_channels(pixbuf) == 4) 
        pixel = pixel | colors[3];
    else
        pixel = pixel | 255;
                 
    return pixel;
}
// 
void test_signal_ico()
{
    printf("Test \"area-updated\" signal on image 'squares32.ico'...\n");
    if(test_signal("squares32.ico", "ico", 300, 50,
                 "ico_initial_pixbuf.bmp", "ico_updated_pixbuf.bmp")
                 )
        printf("Cannot verify parameters if signal was emitting not once.\n");
    else
    {
        printf("Pixel at (0,30) was %x and  become %x.\n", 
            get_pixel("ico_initial_pixbuf.bmp", 0, 30),
            get_pixel("ico_updated_pixbuf.bmp", 0, 30));
    }

}

void test_signal_ani()
{
    printf("Test \"area-updated\" signal on animation image 'hand.ani'(check static image)...\n");
    if(test_signal("hand.ani", "ani", 500, 20,
                 "ani_initial_pixbuf.bmp", "ani_updated_pixbuf.bmp")
                 )
    {
        printf("Cannot verify parameters if signal was emitting not once.\n");
    }
    else
    {
        printf("Pixel at (6,21) was %x and  become %x.\n", 
            get_pixel("ani_initial_pixbuf.bmp", 6, 21),
            get_pixel("ani_updated_pixbuf.bmp", 6, 21));
    }
     
}

void test_signal_gif()
{
    printf("Test \"area-updated\" signal on animation image 'aero.gif'(check static image)...\n");
    if(test_signal("aero.gif", "gif", 300, 200,
                 "gif_initial_pixbuf.bmp", "gif_updated_pixbuf.bmp")
                 )
        printf("Cannot verify parameters if signal was emitting not once.\n");
    else
    {
        printf("Pixel at (0,98) was %x and  become %x.\n", 
            get_pixel("gif_initial_pixbuf.bmp", 0, 98),
            get_pixel("gif_updated_pixbuf.bmp", 0, 98));
    }
}

int main()
{
    g_type_init();
    printf("=================================================================\n");
    test_signal_ico();
    printf("=================================================================\n");
    test_signal_ani();
    printf("=================================================================\n");
    test_signal_gif();
    printf("=================================================================\n");
    return 0;
}